/* FILE: decode.c               Copyright (c), Symmetric Research, 1998
 *
 * DSP32C binary instructon to ASCII mnemonic decoder/disassembly.
 *
 *
 *
 */

#include <stdio.h>
#include <io.h>
#include <math.h>
#include <ctype.h>
#include <malloc.h>
#include <string.h>

#include "decode.h"









/* Static copies of the value to decode and line buffer:
 *
 * Copies of the binary value to decode and line buffer pointers are
 * kept here for easy access by the helper functions below.
 *
 *
 *
 */

#define MAXLINE         64


static unsigned long

          value;

static char

        *line, *l;







/* Format helpers:
 *
 * These functions help construct parts of the final DSP32C decoded
 * instruction.
 *
 *
 */

static

    char illegal[] = "??";



static void format( char *fmt, ... /* char *arg1, *arg2, ... */ ) {

        char *line_start, **stack, *arg, c;



        /* Add the arguments to the output line. */

        line_start = line;
             stack = &fmt;


          while ( (c = *fmt++) != '\0' ) {


                /* Check the fmt string for a % argument. */

                if ( c == '%' ) {


                        if ( (arg = *++stack) == illegal )

                              { *line_start = '\0'; return; }


                        else
                                { while ( *arg ) *line++ = *arg++; }


                        }


                /* If not an arg, add the fmt char to the output line. */

                else
                        *line++ = c;

                }


        *line = '\0';


}
static char *hex16( void) {

        static char buffer[8];

            unsigned N;



        /* Decode a 16 bit immediate value. */

        N = value&0xFFFF;

           sprintf( buffer, "0x%04X", N);


        return( buffer);


}
static char *hex21( void) {

        static char buffer[16];

            unsigned long N;


        /* Decode a 21 bit immediate value. */

        N = value&0x1FFFFFL;

            sprintf( buffer, "0x%06lX", N);


        return( buffer);


}
static char *hex24( void) {

        static char buffer[16];

            unsigned long M;



        /* Decode a 24 bit immediate value. */

        M = ((value>>5) & 0x00FF0000L) + (value & 0x0000FFFFL);

            sprintf( buffer, "0x%06lX", M);


        return( buffer);


}
static char *reg_name[] = {

        illegal,                /* 00   */

        "r1",                   /* 01   */
        "r2",                   /* 02   */
        "r3",                   /* 03   */
        "r4",                   /* 04   */
        "r5",                   /* 05   */
        "r6",                   /* 06   */
        "r7",                   /* 07   */
        "r8",                   /* 08   */
        "r9",                   /* 09   */
        "r10",                  /* 0A   */
        "r11",                  /* 0B   */
        "r12",                  /* 0C   */
        "r13",                  /* 0D   */
        "r14",                  /* 0E   */

        illegal,                /* 0F   */
        illegal,                /* 10   */

        "r15",                  /* 11   */
        "r16",                  /* 12   */
        "r17",                  /* 13   */
        "r18",                  /* 14   */
        "r19",                  /* 15   */

        illegal,                /* 16   */
        illegal,                /* 17   */

        "r20",                  /* 18   */
        "r21",                  /* 19   */

        illegal,                /* 1A   */
        illegal,                /* 1B   */
        illegal,                /* 1C   */

        "r22",                  /* 1D   */

        illegal,                /* 1E   */
        illegal,                /* 1F   */

        };


static char *regN( void) {


        /* Decode an rN register in the normal position. */

        return( reg_name[ (value>>16)&0x1F]);


}
static char *regH( void) {


        /* Decode the pc register along with the normal rN registers. */

        if ( (value & 0x001F0000L) == 0x000F0000L )

                return( "pc");

        else
                return( regN());


}
static char *regM( void) {


        /* Decode an rN register in the M field position. */

        return( reg_name[ (value>>21)&0x1F]);



}
static char *regH_N( void) {

        static char buffer[16];

            char *r;



        /* Decode rH+N with 16 bit N. */

        if ( (r = regH()) != illegal ) {


                if ( value & 0x0000FFFFL )

                        sprintf( buffer, "%s + %s", r, hex16());

                else
                        sprintf( buffer, "%s", regH());



                return( buffer);

                }


        else if ( (value & 0x001F0000L) == 0 )

                return( hex16());


        else

                return( illegal);


}
static char *ninst( void) {

        static char buffer[8];

            unsigned n;



        /* Decode do loop ninst count. */

        n = (value>>16)&0x1F;

            sprintf( buffer, "0x%02X", n);


        return( buffer);


}
static char *nrepeat( void) {

        static char buffer[8];

            unsigned n;



        /* Decode do loop nrepeat count. */
        
        n = value&0x7FF;

            sprintf( buffer, "0x%03X", n);


        return( buffer);


}
static char *regS( void) {


        /* Decode an rN reigster in the S field position. */

        return( reg_name[ (value>>5)&0x1F]);


}
static char *regT( void) {


        /* Decode an rN register in the T field position. */

        return( reg_name[ value&0x1F]);


}
static char *size( void) {

        static char *size_table[] = {

                "h",    /* 0  high byte    */
                "l",    /* 1  low byte     */
                "",     /* 2  word         */
                "e",    /* 3  long         */

                };



        /* Decode the size field for the data move group. */

        return( size_table[ (value>>22)&0x3]);


}
static char *ioreg_name[] = {

        illegal,                /* 00   */
        illegal,                /* 01   */
        illegal,                /* 02   */
        illegal,                /* 03   */

        "ibuf",                 /* 04   */
        "obuf",                 /* 05   */
        "pdr",                  /* 06   */

        illegal,                /* 07   */
        illegal,                /* 08   */
        illegal,                /* 09   */
        illegal,                /* 0A   */
        illegal,                /* 0B   */
        illegal,                /* 0C   */
        illegal,                /* 0D   */

        "piop",                 /* 0E   */

        illegal,                /* 0F   */
        illegal,                /* 10   */
        illegal,                /* 11   */
        illegal,                /* 12   */
        illegal,                /* 13   */

        "pdr2",                 /* 14   */

        illegal,                /* 15   */

        "pir",                  /* 16   */

        illegal,                /* 17   */
        illegal,                /* 18   */
        illegal,                /* 19   */
        illegal,                /* 1A   */
        illegal,                /* 1B   */
        illegal,                /* 1C   */
        illegal,                /* 1D   */

        "pcw",                  /* 1E   */

        illegal,                /* 1F   */

        };


static char *ioreg_7C( void) {


        /* Decode an io register in bit position 0. */

        return( ioreg_name[ value&0x1F]);


}
static char *ioreg_7D( void) {


        /* Decode an io register in bit position 16. */

        return( ioreg_name[ (value>>16)&0x1F]);


}
static char *mem( void) {

        static char buffer[16];

            char *rP, *rI, *rev; int I;



        /* Decode an indirect memory reference. */

        if ( (rP = reg_name[ (value>>5)&0x1F]) != illegal ) {


            I = value&0x1F; rev = ( (value>>10)&0x1 ? "r" : "" );


#define NONE            0x10
#define DEC             0x16
#define INC             0x17


                if ( I == NONE )

                        sprintf( buffer, "*%s", rP);


                else if ( I == DEC )

                        sprintf( buffer, "*%s--%s", rP, rev);


                else if ( I == INC )

                        sprintf( buffer, "*%s++%s", rP, rev);


                else if ( (rI = reg_name[I]) != illegal )

                        sprintf( buffer, "*%s++%s%s", rP, rI, rev);


                else
                        return( illegal);



                return( buffer);


                }


        else

                return( illegal);



}
static char *cond[] = {

        illegal,        /* 0x00 = 0x00 << 1 | 0  */
        illegal,        /* 0x01 = 0x00 << 1 | 1  */
        "pl",           /* 0x02 = 0x01 << 1 | 0  */
        "mi",           /* 0x03 = 0x01 << 1 | 1  */
        "ne",           /* 0x04 = 0x02 << 1 | 0  */
        "eq",           /* 0x05 = 0x02 << 1 | 1  */
        "vc",           /* 0x06 = 0x03 << 1 | 0  */
        "vs",           /* 0x07 = 0x03 << 1 | 1  */
        "cc",           /* 0x08 = 0x04 << 1 | 0  */
        "cs",           /* 0x09 = 0x04 << 1 | 1  */
        "ge",           /* 0x0A = 0x05 << 1 | 0  */
        "lt",           /* 0x0B = 0x05 << 1 | 1  */
        "gt",           /* 0x0C = 0x06 << 1 | 0  */
        "le",           /* 0x0D = 0x06 << 1 | 1  */
        "hi",           /* 0x0E = 0x07 << 1 | 0  */
        "ls",           /* 0x0F = 0x07 << 1 | 1  */

        "auc",          /* 0x10 = 0x08 << 1 | 0  */
        "aus",          /* 0x11 = 0x08 << 1 | 1  */
        "age",          /* 0x12 = 0x09 << 1 | 0  */
        "alt",          /* 0x13 = 0x09 << 1 | 1  */
        "ane",          /* 0x14 = 0x0A << 1 | 0  */
        "aeq",          /* 0x15 = 0x0A << 1 | 1  */
        "avc",          /* 0x16 = 0x0B << 1 | 0  */
        "avs",          /* 0x17 = 0x0B << 1 | 1  */
        "agt",          /* 0x18 = 0x0C << 1 | 0  */
        "ale",          /* 0x19 = 0x0C << 1 | 1  */
        illegal,        /* 0x1A = 0x0D << 1 | 0  */
        illegal,        /* 0x1B = 0x0D << 1 | 1  */
        illegal,        /* 0x1C = 0x0E << 1 | 0  */
        illegal,        /* 0x1D = 0x0E << 1 | 1  */
        illegal,        /* 0x1E = 0x0F << 1 | 0  */
        illegal,        /* 0x1F = 0x0F << 1 | 1  */

        "ibe",          /* 0x20 = 0x10 << 1 | 0  */
        "ibf",          /* 0x21 = 0x10 << 1 | 1  */
        "obf",          /* 0x22 = 0x11 << 1 | 0  */
        "obe",          /* 0x23 = 0x11 << 1 | 1  */
        "pde",          /* 0x24 = 0x12 << 1 | 0  */
        "pdf",          /* 0x25 = 0x12 << 1 | 1  */
        "pie",          /* 0x26 = 0x13 << 1 | 0  */
        "pif",          /* 0x27 = 0x13 << 1 | 1  */
        "syc",          /* 0x28 = 0x14 << 1 | 0  */
        "sys",          /* 0x29 = 0x14 << 1 | 1  */
        "fbc",          /* 0x2A = 0x15 << 1 | 0  */
        "fbs",          /* 0x2B = 0x15 << 1 | 1  */
        "ireq1_lo",     /* 0x2C = 0x16 << 1 | 0  */
        "ireq1_hi",     /* 0x2D = 0x16 << 1 | 1  */
        "ireq2_lo",     /* 0x2E = 0x17 << 1 | 0  */ 
        "ireq2_hi",     /* 0x2F = 0x17 << 1 | 1  */

        illegal,        /* 0x30 = 0x18 << 1 | 0  */
        illegal,        /* 0x31 = 0x18 << 1 | 1  */
        illegal,        /* 0x32 = 0x19 << 1 | 0  */
        illegal,        /* 0x33 = 0x19 << 1 | 1  */
        illegal,        /* 0x34 = 0x1A << 1 | 0  */
        illegal,        /* 0x35 = 0x1A << 1 | 1  */
        illegal,        /* 0x36 = 0x1B << 1 | 0  */
        illegal,        /* 0x37 = 0x1B << 1 | 1  */
        illegal,        /* 0x38 = 0x1C << 1 | 0  */
        illegal,        /* 0x39 = 0x1C << 1 | 1  */
        illegal,        /* 0x3A = 0x1D << 1 | 0  */
        illegal,        /* 0x3B = 0x1D << 1 | 1  */
        illegal,        /* 0x3C = 0x1E << 1 | 0  */
        illegal,        /* 0x3D = 0x1E << 1 | 1  */
        illegal,        /* 0x3E = 0x1F << 1 | 0  */ 
        illegal,        /* 0x3F = 0x1F << 1 | 1  */

        };


static char *cacond_6AB( void) {

        static char buffer[16];

            char *c;



        /* Decode a cacond in bit position 12 for CAU_6AB instructions. */

        if ( value & 0x00000400L ) {


                if ( (c = cond[ ( value>>12)&0xF]) != illegal ) {

                        sprintf( buffer, "if ( %s ) ", c);

                            return( buffer);

                        }

                else
                        return( illegal);

                }


        else

                return( "");

}










/* CAU formatting:
 *
 * These functions format DSP32C CAU instructions.
 *
 *
 *
 */

static void FMT_unknown( void ) {


        /* Format an unknown instruction. */

        sprintf( line, "");



}
static void CAU_1( void) {
        
        unsigned c, rH, N;



        /* Format a cond branch: if ( cond ) goto {rH,N,rH+N}. */

        if ( value == 0x0000000L )

                format( "nop");


        else {


            c = (value>>21)&0x3F;


                if ( c == 0x01 ) {

                        rH = (value>>16)&0x1F; N = value&0xFFFF;


                        if ( rH == 0x1E && N == 0x0000 )

                                format( "ireturn");


                        else if ( N == 0x0000 )

                                format( "return (%)", regN());


                        else
                                FMT_unknown();


                        }


                else

                        format( "if ( % ) goto %", cond[c], regH_N());

                }



}
static void CAU_3A( void) {


        /* Format the loop counter: if (rM-- >= 0) goto {rH,N,rH+N}. */

        format( "if ( %-- >= 0 ) goto %", regM(), regH_N());


}
static void CAU_3B( void) {


        /* Format do immediate: do J,K. */

        format( "do %, %", ninst(), nrepeat());


}
static void CAU_3C( void) {


        /* Format do register: do J,rH. */

        format( "do %, %", ninst(), regT());


}
static void CAU_4( void) {


        /* Format a call: call {rH,N,rH+N} (rM). */

        format( "call % (%)", regH_N(), regM());


}
static void CAU_5A( void) {

        char *r; unsigned rM, rH;



        /* Format 16 bit three op immediate add: rM = rH+N. */

        if ( (r = regM()) != illegal )

                format( "% = %", r, regH_N());


        else {


            /* special case {ioc,dauc} = N ... */

              rM = (value>>21)&0x1F;
              rH = (value>>16)&0x1F;


                if ( rM == 0x1B )

                        format( "ioc = %", hex21());


                else if ( rM == 0x1A && rH == 0x00 )

                        format( "dauc = %", hex16());


                else
                        FMT_unknown();

                }



}
static void CAU_5B( void) {


        /* Format 24 bit three op immediate add: rMe = rH+N. */

        format( "%e = %", regM(), regH_N());



}
static void CAU_6AB( void) {


    static struct func_table { char *op; int fmt; }

        *f, func[] = {

                "+",        1,     /* 0x0   rD[e] = rT + rS     */
                "* 2",      3,     /* 0x1   rD[e] = rS * 2      */
                "-",        2,     /* 0x2   rD[e] = rS - rT     */
                "#",        1,     /* 0x3   rD[e] = rT # rS     */
                "-",        1,     /* 0x4   rD[e] = rT - rS     */
                "-",        4,     /* 0x5   rD[e] = - rS        */
                "&~",       1,     /* 0x6   rD[e] = rT &~ rS    */
                "-",        5,     /* 0x7   rD[e] - rS          */
                "^",        1,     /* 0x8   rD[e] = rT ^ rS     */
                ">>> 1",    3,     /* 0x9   rD[e] = rS >>> 1    */
                "|",        1,     /* 0xA   rD[e] = rT | rS     */
                "<<< 1",    3,     /* 0xB   rD[e] = rS <<< 1    */
                ">> 1",     3,     /* 0xC   rD[e] = rS >> 1     */
                "/ 2",      3,     /* 0xD   rD[e] = rS / 2      */
                "&",        1,     /* 0xE   rD[e] = rT & rS     */
                "&",        5,     /* 0xF   rD[e] & rS          */

                                   /* where if dyadic, rT = rD ... */
                };


            char *rD, *size, *rT, *op, *rS, *c;




        /* Format 16/24 bit reg arith: [if (ca cond)] rD[e] = rT op rS. */

        rD = regN();
        rS = regS();

          rT = ( value & 0x00000800L ? regT() : rD );

            size = ( value & 0x80000000L ? "e" : "");

                f = &func[ (value>>21)&0xF]; op = f->op;



            /* get the possible conditional execution ... */

                c = cacond_6AB();





        /* Format the instruction, peeling off some special cases. */

        if ( (value & 0x7FE00BE0L) == 0x18000A00L )

                /* special case rD = rT ... */

                format( "%%% = %", c, rD, size, rT);



        else if ( (value & 0x7FE00BFFL) == 0x180002E0L )

                /* special case rD++ ... */

                format( "%%%++", c, rD, size);



        else if ( (value & 0x7FE00BFFL) == 0x188002E0L )

                /* special case rD-- ... */

                format( "%%%--", c, rD, size);



        else {


            /* format a normal CAU_6AB instruction ... */


                switch ( f->fmt ) {

                        case 1:

                            format( "%%% = % % %", c, rD, size, rT, op, rS);
                            break;

                        case 2:

                            format( "%%% = % % %", c, rD, size, rS, op, rT);
                            break;

                        case 3:

                            format( "%%% = % %", c, rD, size, rS, op);
                            break;

                        case 4:

                            format( "%%% = % %", c, rD, size, op, rS);
                            break;

                        case 5:

                            format( "%%% % %", c, rD, size, op, rS);
                            break;


                        default:

                            FMT_unknown();
                            break;

                        }

                }



}
static void CAU_6CD( void) {

    static struct func_table { char *op; int fmt; }

        *f, func[] = {

                illegal,    0,     /* 0x0                       */
                illegal,    0,     /* 0x1                       */
                "-",        2,     /* 0x2  rD[e] = N - rD       */
                "#",        1,     /* 0x3  rD[e] = rD # N       */
                "-",        1,     /* 0x4  rD[e] = rD - N       */
                illegal,    0,     /* 0x5                       */
                "&~",       1,     /* 0x6  rD[e] = rD &~ N      */
                "-",        3,     /* 0x7  rD[e] - N            */
                "^",        1,     /* 0x8  rD[e] = rD ^ N       */
                illegal,    0,     /* 0x9                       */
                "|",        1,     /* 0xA  rD[e] = rD | N       */
                illegal,    0,     /* 0xB                       */
                illegal,    0,     /* 0xC                       */
                illegal,    0,     /* 0xD                       */
                "&",        1,     /* 0xE  rD[e] = rD & N       */
                "&",        3,     /* 0xF  rD[e] & N            */

                                   /* where only dyadic is available ... */
                };


          char *rD, *size, *op, *N;




        /* Format 16/24 bit immediate arith: rD[e] = rD op N. */

        rD = regN();

          size = ( value & 0x80000000L ? "e" : "");

                f = &func[ (value>>21)&0xF]; op = f->op; N = hex16();




        /* Format the instruction. */

        switch ( f->fmt ) {


                case 1:

                        format( "%% %= %", rD, size, op, N);
                        break;

                case 2:

                        format( "%% = % % %", rD, size, N, op, rD);
                        break;

                case 3:

                        format( "%% % %", rD, size, op, N);
                        break;

                default:

                        FMT_unknown();
                        break;

                }



}

#define dir     (value & 0x01000000L)   /* test the direction bit */

static void CAU_7A( void) {


        /* Format a direct memory move: rD[size] <-> *N. */

        if ( dir )

                format( "*% = %%", hex16(), regH(), size());

        else

                format( "%% = *%", regH(), size(), hex16());


}
static void CAU_7B( void) {


        /* Format an indirect memory move: rD[size] <-> *rP++rIr. */

        if ( dir )

                format( "% = %%", mem(), regH(), size());

        else
                format( "%% = %", regH(), size(), mem());



}
static void CAU_7C( void) {


        /* Format an IO to CAU move: io <-> rD[size]. */
        
        if ( dir )

                format( "% = %%", ioreg_7C(), regH(), size());

        else
                format( "%% = %", regH(), size(), ioreg_7C());



}
static void CAU_7D( void) {


        /* Format an IO to indirect memory move: io[size] <-> *rP++rIr. */

        if ( dir )

                format( "% = %%", mem(), ioreg_7D(), size());

        else
                format( "%% = %", ioreg_7D(), size(), mem());


}
static void CAU_8A( void) {


        /* Format a goto with 24 bit offset: goto rH+M. */

        if ( value & 0x001F0000L ) {


                if ( value & 0x1FE0FFFF )

                        format( "goto % + %", regH(), hex24());

                else
                        format( "goto %", regH());


                }


        else

                format( "goto %", hex24());


}
static void CAU_8B( void) {


        /* Format a 24 bit immediate load: rDe = M. */

        format( "%e = %", regN(), hex24());


}
static void CAU_8C( void) {


        /* Format call immediate: call M (rN). */

        format( "call % (%)", hex24(), regN());


}








/* DAU formatting:
 *
 * These functions format DSP32C DAU floating point instructions.
 *
 *
 *
 * NOTE:
 *
 * The special cases of Y and Z pbits = 1111 are not completely
 * documented in the AT&T DSP32C Information manual.  Page A-11 says
 * this value is not allowed.  It is in fact used when the pbits of the
 * Y and X or Z and Y fields would be the same for the natural encoding.
 * 
 * This encoding is refered to in the old AT&T DSP32 (no C) Information
 * manual, and is followed by the AT&T DSP32C assembler.
 *
 *
 *
 */


#ifdef DAU_NOTE


   /**** old code, cover these special cases ... */

        G = (v>>23)&0xF;
        N = (v>>21)&0x3;
        Y = (v>>7)&0x7F;
        Z = v&0x7F;

        r = 0;                  /* bit reversed inc not possible */


        /* As above, special case Y and Z pbits = 1111 */

        if ( (Z&PBITS) == PBITS )
                
                { Z &= ~PBITS; Z |= Y&PBITS; }


        da_format( "ZN = G( Y)");


#endif






static char

    zero[] = "0.0",
     one[] = "1.0",

        *acc_name[] = {

                "a0",       /* 0x0  */
                "a1",       /* 0x1  */
                "a2",       /* 0x2  */
                "a3",       /* 0x3  */

                zero,       /* 0x4  */
                one,        /* 0x5  */

                illegal,    /* 0x6  */
                illegal,    /* 0x7  */

                };



static char *aN( void) {


        /* Decode an accumulator in the aN field position. */

        return( acc_name[ (value>>21)&0x3]);


}
static char *aM( void) {


        /* Decode an accumulator in the aM field position. */

        return( acc_name[ (value>>26)&0x7]);



}
static char *regP[] = {

        illegal,        /* 0x00  */
        "r1",           /* 0x01  */
        "r2",           /* 0x02  */
        "r3",           /* 0x03  */
        "r4",           /* 0x04  */
        "r5",           /* 0x05  */
        "r6",           /* 0x06  */
        "r7",           /* 0x07  */
        "r8",           /* 0x08  */
        "r9",           /* 0x09  */
        "r10",          /* 0x0A  */
        "r11",          /* 0x0B  */
        "r12",          /* 0x0C  */
        "r13",          /* 0x0D  */
        "r14",          /* 0x0E  */
        illegal,        /* 0x0F  */

        };

static char *regI[] = {

        "",             /* 0x00  */
        "++r15",        /* 0x01  */
        "++r16",        /* 0x02  */
        "++r17",        /* 0x03  */
        "++r18",        /* 0x04  */
        "++r19",        /* 0x05  */
        "--",           /* 0x06  */
        "++",           /* 0x07  */

        };



#define XFIELD          14            /* XYZ bit field positions */
#define YFIELD          7
#define ZFIELD          0



static char *X( void) {

        static char buffer[16];

            unsigned p, i;


        static char *regI_direct[] = {  /* X field direct addressing */

                "a0",           /* 0x00  */
                "a1",           /* 0x01  */
                "a2",           /* 0x02  */
                "a3",           /* 0x03  */

                "ibuf",         /* 0x04  */

                illegal,        /* 0x05  */
                illegal,        /* 0x06  */
                illegal,        /* 0x07  */

                };


        /* Decode a DAU X field memory reference. */

        p = ( value >> (XFIELD+3) )&0xF; i = (value>>XFIELD)&0x7;



                
        /* Special case p = 0xF ... */

        if ( p == 0xF )

                return( illegal);




        /* Format the X field. */

        if ( p ) {

                sprintf( buffer, "*%s%s", regP[p], regI[i]);

                    return( buffer);

                }

        else

              return( regI_direct[i]);



}
static char *Y( void) {

        static char buffer[16];

            unsigned p, i;


        static char *regI_direct[] = {  /* Y field direct addressing */

                "a0",           /* 0x00  */
                "a1",           /* 0x01  */
                "a2",           /* 0x02  */
                "a3",           /* 0x03  */

                "ibuf",         /* 0x04  */

                illegal,        /* 0x05  */
                illegal,        /* 0x06  */
                illegal,        /* 0x07  */

                };


        /* Decode a DAU Y field memory reference. */

        p = ( value>>(YFIELD+3) )&0xF; i = (value>>YFIELD)&0x7;





        /* Special case p = 0xF with Y pbits = X pbits ... */

        if ( p == 0xF ) {

                p = ( value>>(XFIELD+3) )&0xF;

                if ( p == 0xF )

                        return( illegal);

                }




        /* Format the Y field. */

        if ( p ) {

                sprintf( buffer, "*%s%s", regP[p], regI[i]);

                    return( buffer);

                }

        else

              return( regI_direct[i]);



}
static char *Z( int rfield) {

        static char buffer[16];
        
            unsigned p, i; char *r;



        static char *regI_direct[] = {  /* Z field direct addressing */

                illegal,        /* 0x00  */
                illegal,        /* 0x01  */
                illegal,        /* 0x02  */
                illegal,        /* 0x03  */
                illegal,        /* 0x04  */

                "obuf = ",      /* 0x05  */
                "pdr = ",       /* 0x06  */

                "",             /* 0x07  */   /* no write, Z field only */

                };



        /* Decode a DAU Z field memory reference. */

        p = ( value>>(ZFIELD+3) )&0xF;
        i = ( value>>(ZFIELD+0) )&0x7;




        /* Special case p = 0xF with Z pbits = Y pbits ... */

        if ( p == 0xF ) {

            p = ( value>>(YFIELD+3) )&0xF;

                if ( p == 0xF ) {

                        p = ( value>>(XFIELD+3) )&0xF;


                        if ( p == 0xF )

                                return( illegal);

                        }
                }



        /* Format the Z field. */

        if ( p ) {

            r = ( rfield && (value & (0x1L<<25)) ? "r" : "" );


                sprintf( buffer, "*%s%s%s = ", regP[p], regI[i], r);

                    return( buffer);

                }


        else

              return( regI_direct[i]);




}
static char *F( void) {


        /* Decode the adder input sign. */

        return( value & (0x1L<<24) ? "- " : "");


}
static char *S( void) {


        /* Decode the multiplier sign. */

        return( value & (0x1L<<23) ? "-" : "+");

}








static void DAU_1( void) {

        char *acc;


        /* Format 1: [Z = ]aN = {+,-} Y {+,-} aM*X. */

        if ( (acc = aM()) == zero )

                format( "%% = %%", Z(1), aN(), F(), Y());


        else if ( acc == one )

                format( "%% = %% % %", Z(1), aN(), F(), Y(), S(), X());

        else
        

            format( "%% = %% % % * %", Z(1), aN(), F(), Y(), S(), acc, X());



}
static void DAU_2( void) {

        char *acc;


        /* Format 2: aN = {+,-} aM {+,-} (Z=Y)*X. */

        if ( (acc = aM()) == zero )

                format("% = % (%%) * %", aN(), S(), Z(1), Y(), X());


        else if ( acc == one )

                FMT_unknown();

        else
                

            format( "% = %% % (%%) * %", aN(), F(), acc, S(), Z(1), Y(), X());




}
static void DAU_3( void) {

        char *acc;


        /* Format 3: [Z = ]aN = {+,-} aM {+,-} Y*X. */

        if ( (acc = aM()) == zero )

                format( "%% = % % * %", Z(1), aN(), S(), Y(), X());


        else if ( acc == one )

                FMT_unknown();

        else


            format( "%% = %% % % * %", Z(1), aN(), F(), aM(), S(), Y(), X());


}
static void DAU_4( void) {


        /* Format 4: aN = {+,-} (Z=Y) {+,-} X. */

        format( "% = %(%%) % %", aN(), F(), Z(1), Y(), S(), X());


}
static void DAU_5( void) {

        static char *special[] = {

                "ic",           /* 0x0  */
                "oc",           /* 0x1  */
                "float",        /* 0x2  */
                "int",          /* 0x3  */
                "round",        /* 0x4  */
                "ifalt",        /* 0x5  */
                "ifaeq",        /* 0x6  */
                "ifagt",        /* 0x7  */
                illegal,        /* 0x8  */
                illegal,        /* 0x9  */
                "float24",      /* 0xA  */
                "int24",        /* 0xB  */
                "ieee",         /* 0xC  */
                "dsp",          /* 0xD  */
                "seed",         /* 0xE  */ 
                illegal,        /* 0xF  */ 

                };



        /* Format 5, special functions, [Z =]aN = special( Y). */

        format( "%% = %( %)", Z(0), aN(), special[(value>>23)&0xF], Y());




}











/* Decoding helper functions:
 *
 * These functions further decode instructions not fully resolved in the
 * main decoding table.
 *
 *
 */

static void Decode_0x06( void) {


        /* Decode instructions with a leading 0x06:
         *
         * 0x06    000110 0FFFFDDDDDCCCGEKTTTTTSSSSS  CAU_6A
         * 0x06    000110 1FFFFDDDDDNNNNNNNNNNNNNNNN  CAU_6C
         *
         *
         * Mask and test values:
         *
         *   0001 100F FFFD DDDD CCCG EKTT TTTS SSSS  CAU_6A
         *   1111 1110 0000 0000 0000 0000 0000 0000  CAU_6A mask = FE000000
         *   0001 1000 0000 0000 0000 0000 0000 0000  CAU_6A test = 18000000
         *
         *   0001 101F FFFD DDDD NNNN NNNN NNNN NNNN  CAU_6C
         *   1111 1110 0000 0000 0000 0000 0000 0000  CAU_6C mask = FE000000
         *   0001 1010 0000 0000 0000 0000 0000 0000  CAU_6C test = 1A000000
         *
         *
         */

        if ( (value & 0xFE000000L) == 0x18000000L )

                CAU_6AB();

        else if ( (value & 0xFE000000L) == 0x1A000000L )

                CAU_6CD();

        else

                FMT_unknown();


}
static void Decode_0x07( void) {


        /* Decode instructions with a leading 0x07:
         *
         * 0x07    000111 0TWW0HHHHHNNNNNNNNNNNNNNNN  CAU_7A
         * 0x07    000111 0TWW1RRRRRxxxxxrPPPPPIIIII  CAU_7D
         * 0x07    000111 1TWW0HHHHHxxxxxx00000RRRRR  CAU_7C
         * 0x07    000111 1TWW1HHHHHxxxxxrPPPPPIIIII  CAU_7B
         *
         *
         * Mask and test values:
         *
         *   0001 110T WW0H HHHH NNNN NNNN NNNN NNNN  CAU_7A
         *   1111 1110 0010 0000 0000 0000 0000 0000  CAU_7A mask = FE200000
         *   0001 1100 0000 0000 0000 0000 0000 0000  CAU_7A test = 1C000000
         *
         *   0001 110T WW1R RRRR xxxx xrPP PPPI IIII  CAU_7D
         *   1111 1110 0010 0000 0000 0000 0000 0000  CAU_7D mask = FE200000
         *   0001 1100 0010 0000 0000 0000 0000 0000  CAU_7D test = 1C200000
         *
         *   0001 111T WW0H HHHH xxxx xx00 000R RRRR  CAU_7C
         *   1111 1110 0010 0000 0000 0011 1110 0000  CAU_7C mask = FE2003E0
         *   0001 1110 0000 0000 0000 0000 0000 0000  CAU_7C test = 1E000000
         *
         *   0001 111T WW1H HHHH xxxx xrPP PPPI IIII  CAU_7B
         *   1111 1110 0010 0000 0000 0000 0000 0000  CAU_7B mask = FE200000
         *   0001 1110 0010 0000 0000 0000 0000 0000  CAU_7B test = 1E200000
         *
         *
         */

        if ( (value & 0xFE200000L) == 0x1C000000L )

                CAU_7A();

        else if ( (value & 0xFE200000L) == 0x1C200000L )

                CAU_7D();

        else if ( (value & 0xFE2003E0L) == 0x1E000000L )

                CAU_7C();

        else if ( (value & 0xFE200000L) == 0x1E200000L )

                CAU_7B();

        else

                FMT_unknown();



}
static void Decode_0x23( void) {


        /* Decode instructions with a leading 0x23:
         *
         * 0x23    100011 00000IIIII00000NNNNNNNNNNN  CAU_3B
         * 0x23    100011 00001IIIII00000000000SSSSS  CAU_3C
         *
         *
         * Mask and test values:
         *
         *   1000 1100 000I IIII 0000 0NNN NNNN NNNN  CAU_3B
         *   1111 1111 1110 0000 1111 1000 0000 0000  CAU_3B mask = FFE0F800
         *   1000 1100 0000 0000 0000 0000 0000 0000  CAU_3B test = 8C000000
         *
         *   1000 1100 001I IIII 0000 0000 000S SSSS  CAU_3C
         *   1111 1111 1110 0000 1111 1111 1110 0000  CAU_3C mask = FFE0FFE0
         *   1000 1100 0010 0000 0000 0000 0000 0000  CAU_3C test = 8C200000
         *
         *
         */

        if ( (value & 0xFFE0F800L) == 0x8C000000L )

                CAU_3B();

        else if ( (value & 0xFFE0FFE0L) == 0x8C200000L )

                CAU_3C();

        else

                FMT_unknown();


}
static void Decode_0x26( void) {


        /* Decode instructions with a leading 0x26:
         *
         * 0x26    100110 0FFFFDDDDDCCCGEKTTTTTSSSSS  CAU_6B
         * 0x26    100110 1FFFFDDDDDNNNNNNNNNNNNNNNN  CAU_6D
         *
         *
         * Mask and test values:
         *
         *   1001 100F FFFD DDDD CCCG EKTT TTTS SSSS  CAU_6B
         *   1111 1110 0000 0000 0000 0000 0000 0000  CAU_6B mask = FE000000
         *   1001 1000 0000 0000 0000 0000 0000 0000  CAU_6B test = 98000000
         *
         *   1001 101F FFFD DDDD NNNN NNNN NNNN NNNN  CAU_6D
         *   1111 1110 0000 0000 0000 0000 0000 0000  CAU_6D mask = FE000000
         *   1001 1010 0000 0000 0000 0000 0000 0000  CAU_6D test = 9A000000
         *
         *
         */

        if ( (value & 0xFE000000L) == 0x98000000L )

                CAU_6AB();

        else if ( (value & 0xFE000000L) == 0x9A000000L )

                CAU_6CD();

        else

                FMT_unknown();

}










/* Decoder:
 *
 * This function decodes DSP32C instructions.
 *
 * The main decoding table is keyed on the top 6 bits of the instruction.
 *
 *
 */

typedef

    void (*INSTRUCTION)( void);



static INSTRUCTION DECODE_TABLE[] = {

        /* 0x00    00000C CCCCCHHHHHNNNNNNNNNNNNNNNN */    CAU_1,
        /* 0x01    00000C CCCCCHHHHHNNNNNNNNNNNNNNNN */    CAU_1,
        /* 0x02    000010 ************************** */    FMT_unknown,
        /* 0x03    000011 MMMMMHHHHHNNNNNNNNNNNNNNNN */    CAU_3A,
        /* 0x04    000100 MMMMMHHHHHNNNNNNNNNNNNNNNN */    CAU_4,
        /* 0x05    000101 DDDDDHHHHHNNNNNNNNNNNNNNNN */    CAU_5A,
        /* 0x06    000110 ************************** */    Decode_0x06,
        /* 0x07    000111 ************************** */    Decode_0x07,
        /* 0x08    001MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_1,
        /* 0x09    001MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_1,
        /* 0x0A    001MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_1,
        /* 0x0B    001MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_1,
        /* 0x0C    001MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_1,
        /* 0x0D    001MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_1,
        /* 0x0E    001110 rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_4,
        /* 0x0F    001MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_1,
        /* 0x10    010MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_2,
        /* 0x11    010MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_2,
        /* 0x12    010MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_2,
        /* 0x13    010MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_2,
        /* 0x14    010MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_2,
        /* 0x15    010MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_2,
        /* 0x16    010MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_2,
        /* 0x17    010MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_2,
        /* 0x18    011MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_3,
        /* 0x19    011MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_3,
        /* 0x1A    011MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_3,
        /* 0x1B    011MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_3,
        /* 0x1C    011MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_3,
        /* 0x1D    011MMM rFSNNXXXXXXXYYYYYYYZZZZZZZ */    DAU_3,
        /* 0x1E    01111G GGGNN0000000YYYYYYYZZZZZZZ */    DAU_5,
        /* 0x1F    01111G GGGNN0000000YYYYYYYZZZZZZZ */    DAU_5,
        /* 0x20    100000 ************************** */    FMT_unknown,
        /* 0x21    100001 ************************** */    FMT_unknown,
        /* 0x22    100010 ************************** */    FMT_unknown,
        /* 0x23    100011 ************************** */    Decode_0x23,
        /* 0x24    100100 ************************** */    FMT_unknown,
        /* 0x25    100101 DDDDDHHHHHNNNNNNNNNNNNNNNN */    CAU_5B,
        /* 0x26    100110 ************************** */    Decode_0x26,
        /* 0x27    100111 ************************** */    FMT_unknown,
        /* 0x28    101EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8A,
        /* 0x29    101EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8A,
        /* 0x2A    101EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8A,
        /* 0x2B    101EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8A,
        /* 0x2C    101EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8A,
        /* 0x2D    101EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8A,
        /* 0x2E    101EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8A,
        /* 0x2F    101EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8A,
        /* 0x30    110EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8B,
        /* 0x31    110EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8B,
        /* 0x32    110EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8B,
        /* 0x33    110EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8B,
        /* 0x34    110EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8B,
        /* 0x35    110EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8B,
        /* 0x36    110EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8B,
        /* 0x37    110EEE EEEEEHHHHHNNNNNNNNNNNNNNNN */    CAU_8B,
        /* 0x38    111EEE EEEEEMMMMMNNNNNNNNNNNNNNNN */    CAU_8C,
        /* 0x39    111EEE EEEEEMMMMMNNNNNNNNNNNNNNNN */    CAU_8C,
        /* 0x3A    111EEE EEEEEMMMMMNNNNNNNNNNNNNNNN */    CAU_8C,
        /* 0x3B    111EEE EEEEEMMMMMNNNNNNNNNNNNNNNN */    CAU_8C,
        /* 0x3C    111EEE EEEEEMMMMMNNNNNNNNNNNNNNNN */    CAU_8C,
        /* 0x3D    111EEE EEEEEMMMMMNNNNNNNNNNNNNNNN */    CAU_8C,
        /* 0x3E    111EEE EEEEEMMMMMNNNNNNNNNNNNNNNN */    CAU_8C,
        /* 0x3F    111EEE EEEEEMMMMMNNNNNNNNNNNNNNNN */    CAU_8C,

        };


void DSP32C_Decode( unsigned long user_value, char *user_line) {

        unsigned index;



        /* Set up static copies of the value, line pointer, etc. */

         value = user_value;
          line = user_line;
             l = user_line;



        /* Run the decoder, indexing off the top 6 bits. */

        index = (value>>26)&0x3F;

            DECODE_TABLE[index]();

}
